@@ -5,7 +5,7 @@
 
 extern void most_reread_file (void);
 extern void most_read_to_line (int);
-extern Window *most_file_visible (char *);
+extern Most_Window_Type *most_file_visible (char *);
 extern void most_user_get_file(void);
 extern int most_read_file_dsc(int);
 extern void most_get_cdir(char *);
