@@ -99,15 +99,37 @@ static void top_of_buffer_cmd(void)
 static void end_of_buffer_cmd(void)
 {
    while (Most_Buf->fd != -1) most_read_file_dsc(-1);
-   most_update_windows(Most_Num_Lines);
+   most_update_windows (-1);
 }
 #ifdef unix
 #include <signal.h>
 #endif
 static void sys_spawn_cmd(void)
 {
-    int update = 0;
+#ifdef __DJGPP__
+   most_message ("Not implemented.", 1);
+#else
+   int update = 0;
+#ifndef VMS
+   static int can_suspend = -1;
    
+   if (can_suspend == -1)
+     {
+	can_suspend = 0;
+# ifdef SIGTSTP
+	if (SIG_DFL == SLsignal (SIGTSTP, SIG_DFL))
+	  {     
+	     can_suspend = 1;
+	  }
+# endif
+     }
+   if (can_suspend == 0)
+     {
+	most_message ("Shell forbids suspension.", 1);
+	return;
+     }
+#endif
+	
    if (Most_Secure_Mode || Most_Captive_Mode) 
      {
 	most_message ("Operation not permitted by this account.", 1);
@@ -125,6 +147,7 @@ static void sys_spawn_cmd(void)
    most_init_tty();
    most_init_display ();
    if (update) most_redraw_display();
+#endif				       /* DJGPP */
 }
 
 static void redraw_cmd(void)
@@ -132,37 +155,54 @@ static void redraw_cmd(void)
     most_redraw_display();
 }
 
+static int read_integer (char *prompt, int *n)
+{   
+   Most_Mini_Buf[0] = 0;
+   
+   if (!most_read_from_minibuffer(prompt, (char *) Most_Mini_Buf))
+     return -1;
+   
+   if (1 != sscanf((char *) Most_Mini_Buf, "%d", n))
+     {
+	most_message ("Expecting integer.", 1);
+	return -1;
+     }
+   
+   Most_Mini_Buf[0] = 0;
+   return 0;
+}
+
+
 static void goto_line_cmd(void)
 {
-    int n = Most_C_Line;
-    if (Most_Digit_Arg != (int *) NULL) n = *Most_Digit_Arg;
-    else
-      {
-          Most_Mini_Buf[0] = 0;
-	 if (!most_read_from_minibuffer("Goto Line: ", (char *) Most_Mini_Buf)) return;
-          sscanf((char *) Most_Mini_Buf,"%d",&n);
-          Most_Mini_Buf[0] = 0;
+   int n;
+   
+   if (Most_Digit_Arg != (int *) NULL) n = *Most_Digit_Arg;
+   else
+     {
+	if (-1 == read_integer ("Goto Line: ", &n))
+	  return;
       }
-    most_update_windows(n);
+    most_update_windows (n);
 }
 
 static void goto_percent_cmd(void)
 {
     unsigned char *pos;
     int n;
-    if (Most_Digit_Arg != (int *) NULL) n = *Most_Digit_Arg;
-    else
-      {
-          Most_Mini_Buf[0] = 0;
-	 if (!most_read_from_minibuffer("Goto Percent: ", (char *) Most_Mini_Buf)) return;
-          sscanf((char *) Most_Mini_Buf,"%d",&n);
-          Most_Mini_Buf[0] = 0;
-      }
-    if (n < 0) n = 0; else if (n > 100) n = 100;
+
+   if (Most_Digit_Arg != (int *) NULL) n = *Most_Digit_Arg;
+   else
+     {
+	if (-1 == read_integer ("Goto Percent: ", &n))
+	  return;
+     }
+
+   if (n < 0) n = 0; else if (n > 100) n = 100;
    if (Most_Buf->fd != -1) most_read_file_dsc(-1);
    pos = Most_Beg + (n * (Most_Eob - Most_Beg))/100;
    n = most_what_line(pos);
-   most_update_windows(n);
+   most_update_windows (n);
 }
 
 
@@ -207,19 +247,18 @@ static void other_window_cmd(void)
 static void find_next_cmd(void)
 {
    int col, line, n = 1;
-   int ofs;
+   unsigned long ofs;
     
    if (Most_Digit_Arg != NULL) n = *Most_Digit_Arg;
-   line = most_search(Most_Curs_Pos + Most_Search_Dir, n, &col);
+   line = most_search ((Most_Beg + Most_Curs_Offset) + Most_Search_Dir, n, &col);
    if (line < 1) return;
    
-   /* note that only offsets are meaningful!! (version 4) */
-   ofs = Most_Curs_Pos - Most_Beg;
+   ofs = Most_Curs_Offset;
    
    if ((line < Most_Win->beg_line) || (line > Most_Win->beg_line + Most_Win->bot - Most_Win->top))
      most_update_window(line);
    
-   Most_Curs_Pos = Most_Beg + ofs;
+   Most_Curs_Offset = ofs;
    
    Most_Curs_Row = line - Most_C_Line + 1;
    Most_Curs_Col = col;
@@ -235,7 +274,7 @@ static void search_cmd(void)
 				 "Search: ",
 #endif
 				 Most_Search_Str) == -1) return;
-   Most_Curs_Pos = Most_C_Pos;
+   Most_Curs_Offset = Most_C_Pos - Most_Beg;
    find_next_cmd();
 }
 
@@ -454,6 +493,7 @@ void most_init_keymaps (void)
    SLkm_define_key ("\033$", (FVOID_STAR) sys_spawn_cmd, Most_Keymap);
    SLkm_define_key ("\033<", (FVOID_STAR) top_of_buffer_cmd, Most_Keymap);
    SLkm_define_key ("\033>", (FVOID_STAR) end_of_buffer_cmd, Most_Keymap);
+#ifndef __DJGPP__
    SLkm_define_key ("\033OA", (FVOID_STAR) previous_line_cmd, Most_Keymap);
    SLkm_define_key ("\033OB", (FVOID_STAR) next_line_cmd, Most_Keymap);
    SLkm_define_key ("\033OC", (FVOID_STAR) page_right_cmd, Most_Keymap);
@@ -490,10 +530,20 @@ void most_init_keymaps (void)
    SLkm_define_key ("\033[B", (FVOID_STAR) next_line_cmd, Most_Keymap);
    SLkm_define_key ("\033[C", (FVOID_STAR) page_right_cmd, Most_Keymap);
    SLkm_define_key ("\033[D", (FVOID_STAR) page_left_cmd, Most_Keymap);
+   SLkm_define_key ("^@", (FVOID_STAR) set_mark_cmd, Most_Keymap);
+#else
+   SLkm_define_key ("^@H", (FVOID_STAR) previous_line_cmd, Most_Keymap);
+   SLkm_define_key ("^@P", (FVOID_STAR) next_line_cmd, Most_Keymap);
+   SLkm_define_key ("^@M", (FVOID_STAR) page_right_cmd, Most_Keymap);
+   SLkm_define_key ("^@K", (FVOID_STAR) page_left_cmd, Most_Keymap);
+   SLkm_define_key ("^@Q", (FVOID_STAR) page_down_cmd, Most_Keymap);
+   SLkm_define_key ("^@I", (FVOID_STAR) page_up_cmd, Most_Keymap);
+   SLkm_define_key ("^@G", (FVOID_STAR) top_of_buffer_cmd, Most_Keymap);
+   SLkm_define_key ("^@O", (FVOID_STAR) end_of_buffer_cmd, Most_Keymap);
+#endif
    SLkm_define_key ("\t", (FVOID_STAR) page_right_cmd, Most_Keymap);
    SLkm_define_key ("^", (FVOID_STAR) previous_line_cmd, Most_Keymap);
    SLkm_define_key ("^?", (FVOID_STAR) page_up_cmd , Most_Keymap);
-   SLkm_define_key ("^@", (FVOID_STAR) set_mark_cmd, Most_Keymap);
    SLkm_define_key ("^D", (FVOID_STAR) page_down_cmd, Most_Keymap);
    SLkm_define_key ("^F", (FVOID_STAR) search_cmd, Most_Keymap);
    SLkm_define_key ("^H", (FVOID_STAR) help_cmd, Most_Keymap);
@@ -568,6 +618,8 @@ static void digit_arg_cmd(void)
    ch = most_getkey();
    while ((ch >= '0') && (ch <= '9'))
      {
+	if (j == 15) return;
+
 	num[j++] = ch;
 	ch = most_getkey();
      }
