@@ -28,18 +28,18 @@
 
 char *Most_Global_Msg = "Press `Q' to quit, `H' for help, and SPACE to scroll.";
 
-Window *Most_Win;
-Window *Most_Top_Win;
+Most_Window_Type *Most_Win;
+Most_Window_Type *Most_Top_Win;
 int Most_Top_Line;		       /* row number of top window */
 int Most_Curs_Row;
 int Most_Curs_Col;
 int Most_Column = 1;
 int Most_Restore_Width_To = 0;
 char Most_Mini_Buf[256];
+unsigned long Most_Curs_Offset;
 
 static int Beep_Mini = 0;
 static int Minibuffer_Selected;
-unsigned char *Most_Curs_Pos;
 
 void most_message(char *what, int how)
 {
@@ -93,7 +93,7 @@ void most_put_message ()
      {
 	SLsmg_refresh ();
 	/* wait half a second */
-	while (SLang_input_pending (5))
+	if (SLang_input_pending (5))
 	  {
 	     (void) SLang_getkey ();
 	     SLang_Error = SLKeyBoard_Quit = 0;
@@ -268,23 +268,33 @@ static int get_scroll(int *line)
    return(n);
 }
 
+static void check_dirty_flag (void)
+{
+   if (Most_Win->dirty_flag == 0) return;
+   
+   Most_Win->dirty_flag = 0;
+   Most_Num_Lines = Most_Win->n_lines = most_count_lines (Most_Beg, Most_Eob);
+}
 
 void most_update_window (int line)
 {
    int save_line, save_col;
-   unsigned char *save_pos;
+   unsigned long save_offset;
    
    most_read_to_line (line);
+   check_dirty_flag ();
    
    if (Most_Column != Most_Win->col)
      {
 	if (Most_Column < 1) Most_Column = 1;
 	if (Most_Column != Most_Win->col)
 	  {
-	     save_pos = Most_Curs_Pos; save_line = Most_Curs_Row; save_col = Most_Curs_Col;
+	     save_offset = Most_Curs_Offset; 
+	     save_line = Most_Curs_Row; 
+	     save_col = Most_Curs_Col;
 	     most_redraw_window ();
 	     most_update_status ();
-	     Most_Win->curs_pos = Most_Curs_Pos = save_pos;
+	     Most_Win->curs_offset = Most_Curs_Offset = save_offset;
 	     Most_Win->curs_line = Most_Curs_Row = save_line;
 	     Most_Win->curs_col = Most_Curs_Col = save_col;
 	  }
@@ -305,8 +315,14 @@ void most_update_window (int line)
 void most_update_windows (int line)
 {
    int dline,flg;
-   Window *w;
-   
+   Most_Window_Type *w;
+
+   if (line == -1)
+     {
+	check_dirty_flag ();
+	line = Most_Num_Lines;
+     }
+
    dline = line - Most_C_Line;
    most_update_window(line);
    if (!Most_Win->lock) return;
@@ -331,13 +347,16 @@ void most_update_windows (int line)
 void most_redraw_window (void)
 {
    int n, r;
-   
+
+   check_dirty_flag ();
+
    n = Most_Win->bot - Most_Win->top;
    
    if ((Most_C_Line + n) > Most_Num_Lines)
      most_goto_line (Most_Num_Lines - n);
    
-   Most_Win->curs_pos = Most_Curs_Pos = Most_Win->beg_pos = Most_C_Pos;
+   Most_Curs_Offset = Most_C_Pos - Most_Beg;
+   Most_Win->curs_offset = Most_Win->top_offset = Most_Curs_Offset;
    Most_Win->beg_line = Most_C_Line;
    Most_Win->col = Most_Column;
    Most_Win->curs_col = Most_Curs_Col = 1;
@@ -360,7 +379,7 @@ void most_redraw_window (void)
      }
    
    
-   Most_C_Pos = Most_Win->beg_pos;
+   Most_C_Pos = Most_Beg + Most_Win->top_offset;
    Most_C_Line = Most_Win->beg_line;
 }
 
@@ -372,7 +391,7 @@ void most_redraw_window (void)
 
 
 /* associates current window with current buffer */
-void most_save_win_flags(Window *w)
+void most_save_win_flags(Most_Window_Type *w)
 {
    w->flags = 0;
    if (Most_V_Opt) w->flags |= MOST_V_OPT;
@@ -387,7 +406,7 @@ void most_save_win_flags(Window *w)
 void most_window_buffer()
 {
    Most_Win->beg_line = Most_C_Line;
-   Most_Win->beg_pos = Most_C_Pos;
+   Most_Win->top_offset = Most_C_Pos - Most_Beg;
    Most_Win->col = Most_Column;
    Most_Win->buf = Most_Buf;
    Most_Selective_Display = 0;
@@ -395,7 +414,7 @@ void most_window_buffer()
    
    Most_Curs_Row = Most_Win->curs_line = 1;
    Most_Curs_Col = Most_Win->curs_col = 1;
-   Most_Curs_Pos = Most_Win->curs_pos = Most_C_Pos;
+   Most_Curs_Offset = Most_Win->curs_offset = Most_Win->top_offset;
 }
 
 
@@ -411,12 +430,12 @@ static void restore_win_flags (void)
 }
 
 
-static Window *make_window(int r1,int r2)
+static Most_Window_Type *make_window(int r1,int r2)
 {
-   Window *neew;
+   Most_Window_Type *neew;
    
-   neew = (Window *) MOSTMALLOC(sizeof(Window));
-   SLMEMSET ((char *) neew, 0, sizeof (Window));
+   neew = (Most_Window_Type *) MOSTMALLOC(sizeof(Most_Window_Type));
+   SLMEMSET ((char *) neew, 0, sizeof (Most_Window_Type));
    neew->status = (char *) MOSTMALLOC(256);
    SLMEMSET (neew->status, 0, SLtt_Screen_Cols + 1);
    neew->col = Most_Column;
@@ -425,6 +444,7 @@ static Window *make_window(int r1,int r2)
    neew->lock = 0;
    neew->buf = NULL;
    most_save_win_flags(neew);
+   neew->dirty_flag = 1;
    return(neew);
 }
 
@@ -474,7 +494,7 @@ static void update_status1 (void)
    static char neew[256];
    int i,ii,r,x,line_p = 60, len;
    int max_len;
-   Buffer *buf = Most_Win->buf;
+   Most_Buffer_Type *buf = Most_Win->buf;
    
    max_len = SLtt_Screen_Cols - 25;
    r = Most_Win->bot + 1;
@@ -534,9 +554,10 @@ static void update_status1 (void)
 
 void most_update_status (void)
 {
+   check_dirty_flag ();
    
    Most_C_Line = Most_Win->beg_line;
-   Most_C_Pos = Most_Win->beg_pos;
+   Most_C_Pos = Most_Beg + Most_Win->top_offset;
    /* most_set_scroll_region(1,SLtt_Screen_Rows); */
    update_status1();
    /* most_set_scroll_region(Most_Win->top,Most_Win->bot); */
@@ -546,7 +567,7 @@ void most_update_status (void)
 /* splits window-- no screen update, does not affect scrolling region */
 int most_split_window (void)
 {
-   Window *new, *old;
+   Most_Window_Type *new, *old;
    int b2,t2,b1, line;
    
    b2 = Most_Win->bot;
@@ -576,29 +597,30 @@ int most_split_window (void)
 void most_two_windows()
 {
    int line;
-   Window *new, *old;
+   Most_Window_Type *nnew, *old;
    if (!most_split_window()) return;
    
    old = Most_Win;
-   new = Most_Win->next;
-   line = new->beg_line;
-   if (line + new->bot - new->top > Most_Num_Lines)
+   nnew = Most_Win->next;
+   line = nnew->beg_line;
+   if (line + nnew->bot - nnew->top > Most_Num_Lines)
      {
 	most_other_window(1);
-          /* since split window left new window undefined... */
-	Most_C_Pos = old->beg_pos;
+          /* since split window left nnew window undefined... */
+	Most_C_Pos = Most_Beg + old->top_offset;
 	Most_C_Line = old->beg_line;
-	if (Most_Num_Lines <= new->bot - new->top + 1)
+	if (Most_Num_Lines <= nnew->bot - nnew->top + 1)
 	  {
-	     Most_C_Line = new->beg_line = 1;
-	     Most_C_Pos = new->beg_pos = Most_Buf->beg;
+	     Most_C_Line = nnew->beg_line = 1;
+	     Most_C_Pos = Most_Beg;
+	     nnew->top_offset = 0;
 	     most_redraw_window();
 	     most_update_status();
 	  }
 	else if (line > Most_Num_Lines)
 	  {
-	     most_goto_line(Most_Num_Lines - new->bot + new->top);
-	     Most_Win->beg_pos = Most_C_Pos;
+	     most_goto_line(Most_Num_Lines - nnew->bot + nnew->top);
+	     Most_Win->top_offset = Most_C_Pos - Most_Beg;
 	     Most_Win->beg_line = Most_C_Line;
 	     most_redraw_window();
 	     most_update_status();
@@ -606,24 +628,24 @@ void most_two_windows()
 	else
 	  {
 	     most_goto_line(line);
-	     Most_Win->beg_pos = Most_C_Pos;
+	     Most_Win->top_offset = Most_C_Pos - Most_Beg;
 	     Most_Win->beg_line = Most_C_Line;
-	     most_update_window(Most_Num_Lines - new->bot + new->top);
+	     most_update_window(Most_Num_Lines - nnew->bot + nnew->top);
 	  }
 	Most_Win->curs_line = 1;
 	Most_Win->curs_col = Most_Column;
-	Most_Win->curs_pos = Most_C_Pos;
+	Most_Win->curs_offset = Most_C_Pos - Most_Beg;
 	most_other_window(-1);
      }
    else
      {
-	Most_Win = new;
+	Most_Win = nnew;
 	(void) most_forward_line(line - old->beg_line);
-	new->beg_pos = Most_C_Pos;
-	new->beg_line = Most_C_Line;
-	new->curs_line = 1;
-	new->curs_col = Most_Column;
-	new->curs_pos = Most_C_Pos;
+	nnew->beg_line = Most_C_Line;
+	nnew->curs_line = 1;
+	nnew->curs_col = Most_Column;
+	nnew->top_offset = Most_C_Pos - Most_Beg;
+	nnew->curs_offset = Most_C_Pos - Most_Beg;
 	most_update_status();
 	Most_Win = old;
      }
@@ -633,7 +655,7 @@ void most_two_windows()
 
 static void delete_other_windows (void)
 {
-   Window *w, *tmp;
+   Most_Window_Type *w, *tmp;
    
    if (Most_Win->next == Most_Win) return;
    w = Most_Win;
@@ -671,26 +693,35 @@ void most_one_window ()
 }
 
 
-void most_set_window (Window *w)
+void most_set_window (Most_Window_Type *w)
 {
    Most_Win = w;
-   Most_C_Pos = Most_Win->beg_pos;
-   Most_Curs_Pos = Most_Win->curs_pos;
+   
+   if (Most_Win->buf == NULL)
+     {
+	Most_Win->buf = Most_Buf;
+	if (Most_Buf == NULL)
+	  return;
+     }
+   
+   most_switch_to_buffer(Most_Win->buf);
+   
+   Most_C_Pos = Most_Beg + Most_Win->top_offset;
+   Most_Curs_Offset = Most_Win->curs_offset;
    Most_Curs_Row = Most_Win->curs_line;
    Most_Curs_Col = Most_Win->curs_col;
    Most_C_Line = Most_Win->beg_line;
    Most_Column = Most_Win->col;
-   Most_Buf = Most_Win->buf;
-   most_switch_to_buffer(Most_Buf);
    restore_win_flags();
+   check_dirty_flag ();
    most_point_cursor ();
 }
 
 void most_other_window(int n)
 {
    if (!n) return;
-   Most_Win->beg_pos = Most_C_Pos;
-   Most_Win->curs_pos = Most_Curs_Pos;
+   Most_Win->top_offset = Most_C_Pos - Most_Beg;
+   Most_Win->curs_offset = Most_Curs_Offset;
    Most_Win->curs_line = Most_Curs_Row;
    Most_Win->curs_col = Most_Curs_Col;
    Most_Win->beg_line = Most_C_Line;
@@ -726,19 +757,19 @@ static void delete_as_top_window (void)
 /* free buffer for this window if no other windows are viewing it. */
 void most_free_window_buffer (void)
 {
-   Window *w;
-   Buffer *b;
+   Most_Window_Type *w;
+   Most_Buffer_Type *b;
    
    w = Most_Win->next;
    
    while(Most_Win != w)
      {
-	if (!strcmp(Most_Win->buf->file,w->buf->file)) return;
+	if (Most_Win->buf == w->buf) return;
 	w = w->next;
      }
    b = w->buf;
    if (b == NULL) return;
-   if (b->fd != -1) close(b->fd);
+   if ((b->fd != -1) && (b->fd != 0)) close(b->fd);
    if (b->beg != NULL) SLFREE(b->beg);
    Most_Win->buf = NULL;
    SLFREE(b);
@@ -746,7 +777,7 @@ void most_free_window_buffer (void)
 
 void most_free_windows (void)
 {
-   Window *w;
+   Most_Window_Type *w;
    
    while (Most_Win != NULL)
      {
@@ -769,7 +800,7 @@ void most_free_windows (void)
 void most_delete_window (void)
 {
    int new_b, old_b;
-   Window *w;
+   Most_Window_Type *w;
    
    w = Most_Win;
    if (Most_Win->next == Most_Win) return;
@@ -785,9 +816,11 @@ void most_delete_window (void)
 	else
 	  delete_as_top_window();
 	
+	SLFREE(w->status);
 	SLFREE(w);
 	return;
      }
+
    old_b = Most_Win->top - 2;
    new_b = Most_Win->bot;
    most_other_window(-1);
@@ -798,13 +831,16 @@ void most_delete_window (void)
    
    Most_Win->next = w->next;
    Most_Win->next->prev = Most_Win;
+   
+   SLFREE (w->status);
    SLFREE(w);
+
    most_update_status();
 }
 
 void most_redraw_display (void)
 {
-   Window *w;
+   Most_Window_Type *w;
    int n,t;
    
    SLsmg_cls ();
@@ -815,14 +851,16 @@ void most_redraw_display (void)
    do
      {
 	Most_Win = Most_Win->next;
-	Most_C_Pos = Most_Win->beg_pos;
+	most_switch_to_buffer(Most_Win->buf);
+	Most_C_Pos = Most_Beg + Most_Win->top_offset;
 	Most_C_Line = Most_Win->beg_line;
 	Most_Column = Most_Win->col;
-	most_switch_to_buffer(Most_Win->buf);
 	
 	restore_win_flags();
 	n = Most_Win->bot - Most_Win->top;
 	
+	check_dirty_flag ();
+	
 	t = Most_Win->top;
 	most_goto_rc(t, 1);
 	most_display_line();
@@ -834,7 +872,7 @@ void most_redraw_display (void)
 	     if (most_forward_line(1)) most_display_line();
 	  }
 	Most_C_Line = Most_Win->beg_line;
-	Most_C_Pos = Most_Win->beg_pos;
+	Most_C_Pos = Most_Beg + Most_Win->top_offset;
 	update_status1();
      }
    while(Most_Win != w);
